# Copyright (c) 2010 Outpost Embedded, LLC
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Outpost Embedded trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Outpost Embedded, LLC.

import os, sys, errno
from subprocess import Popen, PIPE, STDOUT
from socket import socket, AF_UNIX, SOCK_STREAM


# FIXME: Is there a good way to know what to expect from subprocesses?
SUBPROCESS_ENCODING = 'utf-8'

PROTOCOL_ENCODING = 'utf-8'

SOCKET_NAME = 'socket'
MOUNT_POINT = 'mnt'
LOG_FILE = 'diskdaemon.log'


class Quit(Exception):
    pass


class DaemonError(Exception):
    code = None
    msg = None

    def __init__(self, msg = None):
        super(DaemonError, self).__init__(msg)
        self.msg = msg

    def __unicode__(self):
        if self.msg is not None:
            return u'%s %s' % (self.code, self.msg)
        return unicode(self.code)

    def __str__(self):
        return unicode(self).encode(PROTOCOL_ENCODING)

    @classmethod
    def FromException(cls, e):
        if isinstance(e, IOError):
            # str(...) gives a better message than unicode(...) with IOErrors.
            # I wish that wasn't the case.
            msg = unicode(str(e), errors = 'replace')
        else:
            msg = unicode(e)
        return cls(msg)


class InvalidCommandError(DaemonError):
    code = u'INVALID-COMMAND'


class DeviceBusyError(DaemonError):
    code = u'DEVICE-BUSY'


class DeviceNotFoundError(DaemonError):
    code = u'DEVICE-NOT-FOUND'


class UnknownError(DaemonError):
    code = u'UNKNOWN'


class LinuxDiskDaemon(object):
    uid = None
    socket_path = None
    mount_point = None

    socket = None
    conn = None

    def __init__(self, uid, socket_path, mount_point):
        self.uid = uid
        self.socket_path = socket_path
        self.mount_point = mount_point

    def start(self):
        try:
            os.unlink(self.socket_path)
        except OSError, e:
            if e.errno != errno.ENOENT:
                raise

        self.socket = socket(AF_UNIX, SOCK_STREAM)
        try:
            os.mkdir(self.mount_point)
            try:
                self.socket.bind(self.socket_path)
                os.chown(self.socket_path, self.uid, -1)
                self.socket.listen(1)
            except:
                os.rmdir(self.mount_point)
                raise
        except:
            self.socket.close()
            raise

    def stop(self):
        self.socket.close()
        try:
            os.unlink('socket')
        except OSError:
            pass

        os.rmdir(self.mount_point)

    def run(self):
        self.conn, addr = self.socket.accept()
        try:
            try:
                while True:
                    self._process_command()
            except Quit:
                pass
        finally:
            self.conn.close()

    def _process_command(self):
        chars = []
        while True:
            ch = self.conn.recv(1)
            if not ch:
                raise Quit()
            if ch == '\n':
                break
            chars.append(ch)

        full_command = ''.join(chars)
        parts = full_command.split(' ', 1)
        len_parts = len(parts)

        if len_parts == 0:
            self._send_error(InvalidCommandError('missing command'))
            return
        elif len_parts == 1:
            command = full_command
            arg = None
        elif len_parts == 2:
            command, arg = parts

        if command.startswith('_') or (not command.isupper()):
            self._send_error(InvalidCommandError('unknown command'))
            return

        try:
            fn = getattr(self, command)
        except AttributeError:
            self._send_error(InvalidCommandError('unknown command'))
            return

        if not callable(fn):
            self._send_error(InvalidCommandError('unknown command'))
            return

        try:
            if arg is not None:
                result = fn(arg)
            else:
                result = fn()
        except KeyboardInterrupt:
            raise
        except DaemonError, e:
            self._send_error(e)
        except Exception, e:
            import traceback
            traceback.print_exc()
            self._send_error(UnknownError.FromException(e))
        else:
            self._send_ok(result)

    def format_message(self, msg):
        return msg.replace('\n', '\\n')

    def _send_error(self, e):
        if not isinstance(e, DaemonError):
            raise AssertionError(e)
        msg = self.format_message(unicode(e).encode(PROTOCOL_ENCODING))
        self.conn.send('ERROR %s\n' % msg)

    def _send_ok(self, result = None):
        if result is not None:
            msg = self.format_message(unicode(result).encode(PROTOCOL_ENCODING))
            self.conn.send('OK %s\n' % msg)
        else:
            self.conn.send('OK\n')

    def GETPID(self):
        return str(os.getpid())

    def SETPATH(self, arg):
        # This is a funny command.  It compensates for the fact that root may
        # not have the same PYTHONPATH as the calling user, so imports might
        # fail.
        sys.path[:] = eval(arg)

    def PARTITION(self, arg):
        from fdisk.device import Device
        from fdisk.mbr import MBRDisk

        try:
            device = Device(arg)
        except IOError, e:
            import traceback
            traceback.print_exc()

            if e.errno == errno.ENOENT:
                raise DeviceNotFoundError.FromException(e)
            raise UnknownError.FromException(e)

        try:
            sector_count = device.get_sector_count()

            # Force 255 heads, 63 sectors no matter what the current partition
            # table looks like.  I'm not sure if this is a good idea or not.
            # We'll see.
            geometry = MBRDisk.calculate_geometry(sector_count, 255, 63)
            mbr_disk = MBRDisk(geometry)

            found = False
            for mbr_bin_path in (
              '/usr/lib/syslinux/mbr.bin',
              '/usr/local/lib/syslinux/mbr.bin',
            ):
                if os.path.exists(mbr_bin_path):
                    found = True
                    break

            if not found:
                raise AssertionError(
                  'could not locate mbr.bin; is syslinux installed?'
                )

            with open(mbr_bin_path, 'rb') as f:
                mbr_disk.boot_record.code.read_from_file(f)

            mbr_disk.add_partition(type = 0x0b)

            try:
                mbr_disk.write_to_device(device)
            except IOError, e:
                import traceback
                traceback.print_exc()

                if e.errno == errno.EBUSY:
                    raise DeviceBusyError.FromException(e)
                raise UnknownError.FromException(e)

        finally:
            device.close()

    def MKFS(self, arg):
        self._exec(['mkfs', '-t', 'vfat', arg])

    def BOOTLOADER(self, arg):
        self._exec(['syslinux', arg])

    def MOUNT(self, arg):
        self._exec([
          'mount',
          '-t', 'vfat',
          '-o', 'uid=%u' % self.uid,
          arg,
          self.mount_point,
        ])
        return self.mount_point

    def UMOUNT(self, arg):
        self._exec(['umount', arg])

    def _exec(self, args):
        popen = Popen(args, stdout = PIPE, stderr = STDOUT)
        try:
            popen.wait()
            output = popen.stdout.read()
        finally:
            popen.stdout.close()
        if popen.returncode != 0:
            msg = output.decode(SUBPROCESS_ENCODING)
            raise UnknownError(msg)


def main(argv):
    if len(argv) != 3:
        print >>sys.stderr, 'need two arguments'
        return 1

    wd = os.path.abspath(argv[1])
    uid = int(argv[2])

    if not os.path.isdir(wd):
        os.mkdir(wd)

    os.chdir(wd)

    log_path = os.path.join(wd, LOG_FILE)

    # Note: It's particularly important that we close stderr.  gksu on
    # Debian/Ubuntu reads stderr over a pipe and copies the data to the
    # terminal.  The result is that gksu will not exit until stderr is closed,
    # despite the forks below.

    os.close(1)
    os.open(
      log_path,
      (os.O_CREAT | os.O_WRONLY | os.O_TRUNC),
      0666,
    )

    os.close(2)
    os.dup(1)

    socket_path = os.path.join(wd, SOCKET_NAME)
    mount_point = os.path.join(wd, MOUNT_POINT)

    daemon = LinuxDiskDaemon(uid, socket_path, mount_point)
    daemon.start()
    try:
        if os.fork():
            os._exit(0)

        os.setsid()

        if os.fork():
            os._exit(0)

        daemon.run()

    finally:
        daemon.stop()

    os.unlink(log_path)
    os.chdir('/')
    os.rmdir(wd)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))
