# Copyright (c) 2010 Outpost Embedded, LLC
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Outpost Embedded trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Outpost Embedded, LLC.

import os, Queue, logging

from PyQt4 import QtCore, QtGui
from tasks import InstructionError

from rapidwriter.dialog import Ui_Dialog
from rapidwriter.tasks_ import PrepareDiskTask
from rapidwriter.util import (
  qt_string_to_unicode,
  QThreadWithEventLoop,
  FunctionObject,
)


logger = logging.getLogger()


# FIXME: Move this to pytasks.qt4.

class TaskObject(QtCore.QObject):
    task_changes_emitted = QtCore.pyqtSignal(list)

    running_changed = QtCore.pyqtSignal((object, bool), (bool,))
    cancelled_changed = QtCore.pyqtSignal((object, bool), (bool,))
    status_changed = QtCore.pyqtSignal((object, basestring), (basestring,))
    progress_changed = QtCore.pyqtSignal((object, int), (int,))
    finished_changed = QtCore.pyqtSignal((object, bool), (bool,))
    completed_changed = QtCore.pyqtSignal((object, bool), (bool,))
    failed_changed = QtCore.pyqtSignal((object, bool), (bool,))
    failure_changed = QtCore.pyqtSignal((object, object), (object,))
    paused_changed = QtCore.pyqtSignal((object, bool), (bool,))

    running_subtask_changed = QtCore.pyqtSignal((object, object), (object,))

    started = QtCore.pyqtSignal()
    cancelled = QtCore.pyqtSignal()
    completed = QtCore.pyqtSignal()
    failed = QtCore.pyqtSignal()
    finished = QtCore.pyqtSignal()
    paused = QtCore.pyqtSignal()
    resumed = QtCore.pyqtSignal()

    task = None
    _instruction_queue = None
    _pending_instructions = None

    def __init__(self, task):
        super(TaskObject, self).__init__()
        self.task = task
        self._instruction_queue = Queue.Queue()
        self._pending_instructions = []

    @QtCore.pyqtSlot()
    def run(self):
        for task_changes in self.task.run():
            self.task_changes_emitted.emit(task_changes)

            for task_change in task_changes:
                if (task_change.parameter == 'running'):
                    self.running_changed[object, bool].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        if task_change.value:
                            self.started.emit()
                        self.running_changed[bool].emit(task_change.value)

                elif (task_change.parameter == 'cancelled'):
                    self.cancelled_changed[object, bool].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        if task_change.value:
                            self.cancelled.emit()
                        self.cancelled_changed[bool].emit(task_change.value)

                elif (task_change.parameter == 'status'):
                    self.status_changed[object, basestring].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        self.status_changed[basestring].emit(task_change.value)

                elif task_change.parameter == 'progress':
                    self.progress_changed[object, int].emit(
                      task_change.task,
                      int(task_change.value),
                    )
                    if task_change.task is self.task:
                        self.progress_changed[int].emit(int(task_change.value))

                elif (task_change.parameter == 'finished'):
                    self.finished_changed[object, bool].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        if task_change.value:
                            self.finished.emit()
                        self.finished_changed[bool].emit(task_change.value)

                elif (task_change.parameter == 'completed'):
                    self.completed_changed[object, bool].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        if task_change.value:
                            self.completed.emit()
                        self.completed_changed[bool].emit(task_change.value)

                elif (task_change.parameter == 'failed'):
                    self.failed_changed[object, bool].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        if task_change.value:
                            self.failed.emit()
                        self.failed_changed[bool].emit(task_change.value)

                elif (task_change.parameter == 'failure'):
                    self.failure_changed[object, object].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        self.failure_changed[object].emit(task_change.value)

                elif task_change.parameter == 'paused':
                    self.paused_changed[object, bool].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        if task_change.value:
                            self.paused.emit()
                        else:
                            self.resumed.emit()
                        self.paused_changed[bool].emit(task_change.value)

                elif task_change.parameter == 'running_subtask':
                    self.running_subtask_changed[object, object].emit(
                      task_change.task,
                      task_change.value,
                    )
                    if task_change.task is self.task:
                        self.running_subtask_changed[object].emit(
                          task_change.value,
                        )

            instruction = None
            try:
                instruction = self._pending_instructions.pop(0)
            except IndexError:
                try:
                    instruction = self._instruction_queue.get(block = False)
                except Queue.Empty:
                    pass

            if instruction is not None:
                try:
                    if instruction == 'cancel':
                        if not self.task.terminal:
                            self.task.cancel()
                    elif instruction == 'pause':
                        if 'paused' not in self.task.final_state:
                            self.task.pause()
                    elif instruction == 'resume':
                        if 'paused' in self.task.final_state:
                            self.task.resume()
                except InstructionError:
                    self._pending_instructions.insert(0, instruction)
                    logger.exception('instruction error: %r', instruction)

    @QtCore.pyqtSlot()
    def cancel(self):
        self._instruction_queue.put('cancel')

    @QtCore.pyqtSlot()
    def pause(self):
        self._instruction_queue.put('pause')

    @QtCore.pyqtSlot()
    def resume(self):
        self._instruction_queue.put('resume')


class RapidWriterDialog(Ui_Dialog, QtGui.QDialog):
    backend = None
    disks = None

    selected_disk = None
    source_file = None
    task_object = None
    task_thread = None

    def __init__(self, backend_cls, source_file = None):
        super(RapidWriterDialog, self).__init__()

        self.disks = []

        self.setupUi(self)

        # Don't allow window resizing.
        self.setMinimumSize(self.size())
        self.setMaximumSize(self.size())

        self.deviceComboBox.currentIndexChanged.connect(
          self.handle_device_index_changed,
        )
        self.sourceFileButton.clicked.connect(self.select_source_file)
        self.writeButton.clicked.connect(self.confirm)
        self.progressBar.setMaximum(100)

        self._refresh_widgets()

        if source_file is not None:
            self.set_source_file(source_file)

        self.backend = backend_cls(self.add_disk, self.remove_disk)

    def exec_(self):
        self.backend.start()
        try:
            return super(RapidWriterDialog, self).exec_()
        finally:
            self.backend.stop()

    def _refresh_widgets(self):
        logger.debug(
          (
            'gui: refreshing widgets: '
            'task_thread = %r, source_file = %r, selected_disk = %r'
          ),
          self.task_thread,
          self.source_file,
          self.selected_disk,
        )
        if (not self.task_thread) and self.source_file and self.selected_disk:
            self.writeButton.setEnabled(True)
        else:
            self.writeButton.setEnabled(False)

        if not self.task_thread:
            self.deviceComboBox.setEnabled(True)
            self.sourceFileButton.setEnabled(True)
        else:
            self.deviceComboBox.setEnabled(False)
            self.sourceFileButton.setEnabled(False)

    def set_source_file(self, source_file):
        self.source_file = source_file
        self.sourceFileLabel.setText(os.path.basename(source_file))
        self._refresh_widgets()

    def add_disk(self, disk):
        logger.info('gui: adding disk: %r', disk)
        def f():
            # Only use removable devices.
            if disk.is_removable():
                self.disks.append(disk)
                self.deviceComboBox.addItem(disk.get_ui_string())
        # We don't know what thread we're being called from ...
        FunctionObject.call_in_main_thread(f)

    def remove_disk(self, disk):
        logger.info('gui: removing disk: %r', disk)
        def f():
            found = False
            for index, disk_ in enumerate(self.disks):
                if disk == disk_:
                    found = True
                    break
            if found:
                del self.disks[index]
                self.deviceComboBox.removeItem(index)
        # We don't know what thread we're being called from ...
        FunctionObject.call_in_main_thread(f)

    @QtCore.pyqtSlot(int)
    def handle_device_index_changed(self, index):
        logger.info('gui: device index changed: %r', index)

        if self.selected_disk is not None:
            del self.selected_disk

        if index >= 0:
            self.selected_disk = self.disks[index]
            logger.info('gui: selected disk is %r', self.selected_disk)

        self._refresh_widgets()

    def closeEvent(self, event):
        if self.task_thread is not None:
            self.task_object.pause()

            reply = QtGui.QMessageBox.warning(
              self,
              'Abort?',
              'Would you like to abort the write process?',
              QtGui.QMessageBox.Yes,
              QtGui.QMessageBox.No,
            )

            self.task_object.resume()
            if reply == QtGui.QMessageBox.Yes:
                self.task_object.cancel()

            event.ignore()

    def confirm(self):
        reply = QtGui.QMessageBox.warning(
          self,
          'Warning',
          'This will destroy all data on the device.  Are you sure you want to do this?',
          QtGui.QMessageBox.Yes,
          QtGui.QMessageBox.No,
        )

        if reply == QtGui.QMessageBox.Yes:
            self.begin()

    # FIXME: Better method name.
    def begin(self):
        if self.selected_disk is None:
            return

        self.progressBar.setValue(0)
        self.statusLabel.setText('')

        self.task_object = TaskObject(
          PrepareDiskTask(self.selected_disk, self.source_file),
        )
        self.task_object.status_changed[object, basestring].connect(
          self._handle_status_changed,
        )
        self.task_object.progress_changed[int].connect(
          self._handle_progress_changed,
        )
        self.task_object.failed.connect(self._handle_task_failed)
        self.task_object.paused.connect(self._handle_task_paused)
        self.task_object.resumed.connect(self._handle_task_resumed)
        self.task_object.cancelled.connect(self._handle_task_cancelled)
        self.task_object.completed.connect(self._handle_task_completed)

        self.task_thread = QThreadWithEventLoop()
        self.task_thread.finished.connect(self.task_thread_ended)
        self.task_thread.terminated.connect(self.task_thread_ended)

        self.task_object.moveToThread(self.task_thread)

        # Start the task when the thread starts.
        self.task_thread.started.connect(self.task_object.run)
        # Quit the thread when the task is finished.
        self.task_object.finished.connect(self.task_thread.quit)

        self.backend.process_thread(self.task_thread)

        self.task_thread.start()

        self._refresh_widgets()

    @QtCore.pyqtSlot(int)
    def _handle_progress_changed(self, value):
        self.progressBar.setValue(value)

    @QtCore.pyqtSlot(object, basestring)
    def _handle_status_changed(self, task, status):
        if not isinstance(task, PrepareDiskTask):
            self.status(unicode(status))

    @QtCore.pyqtSlot()
    def _handle_task_failed(self):
        logger.error('FAILED')
        logger.error(unicode(self.task_object.task.failure))
        QtGui.QMessageBox.critical(
          self,
          'Failed',
          (
            u'Creating the deployment disk failed:\n\n%s'
            % unicode(self.task_object.task.failure)
          ),
          QtGui.QMessageBox.Ok,
        )

    @QtCore.pyqtSlot()
    def _handle_task_paused(self):
        logger.info('PAUSED')

    @QtCore.pyqtSlot()
    def _handle_task_resumed(self):
        logger.info('RESUMED')

    @QtCore.pyqtSlot()
    def _handle_task_cancelled(self):
        logger.info('CANCELLED')
        self.done(0)

    @QtCore.pyqtSlot()
    def _handle_task_completed(self):
        logger.info('DONE')
        self.status('Done!')

    @QtCore.pyqtSlot(basestring)
    def status(self, text):
        self.statusLabel.setText(text)

    @QtCore.pyqtSlot()
    def task_thread_ended(self):
        del self.task_thread
        del self.task_object
        self._refresh_widgets()

    def select_source_file(self):
        source_file = QtGui.QFileDialog.getOpenFileName(
          self,
          'Select Build Archive',
          '.',
          'Build Archive (*.bar)',
        )
        if source_file:
            self.set_source_file(qt_string_to_unicode(source_file))
