#!/usr/bin/env python

# Author: Forest Bond <forest.bond@rapidrollout.com>
# This file is in the public domain.

import sys, os, commands
from urllib import urlopen
from subprocess import check_call
from distutils.cmd import Command
from distutils import log

from setuplib import setup, py2exe as _py2exe


extra_kwargs = {'cmdclass': {}}


###


# FIXME: clean should remove the generated UI files.

class genui(Command):
    description = 'generate UI files'
    user_options = []

    def initialize_options(self):
        self.ui_dest_dir = None

    def finalize_options(self):
        self.ui_dest_dir = 'modules'

    def run(self):
        resources_module_file = os.path.join(
          self.ui_dest_dir,
          'rapidwriter',
          'resources_rc.py',
        )
        qrc_file = os.path.join(
          os.path.dirname(__file__),
          'ui',
          'resources.qrc',
        )

        dialog_module_file = os.path.join(
          self.ui_dest_dir,
          'rapidwriter',
          'dialog.py',
        )
        ui_file = os.path.join(os.path.dirname(__file__), 'ui', 'dialog.ui')

        self.mkpath(os.path.dirname(resources_module_file))
        self.mkpath(os.path.dirname(dialog_module_file))

        # Note: Must use shell on win32 to perform PATH searching.
        # See CreateProcess documentation for details.
        log.info('creating %s', resources_module_file)
        check_call(
          ['pyrcc4', qrc_file, '-o', resources_module_file],
          shell = (sys.platform == 'win32'),
        )
        log.info('creating %s', dialog_module_file)
        check_call(
          ['pyuic4', ui_file, '-o', dialog_module_file],
          shell = (sys.platform == 'win32'),
        )


extra_kwargs['cmdclass']['genui'] = genui


###


from distutils.command.build_py import build_py as _build_py


class build_py(_build_py):
    def run(self):
        genui_obj = self.get_finalized_command('genui')
        genui_obj.ui_dest_dir = self.build_lib
        genui_obj.run()
        return _build_py.run(self)


extra_kwargs['cmdclass']['build_py'] = build_py


###


if _py2exe is None:
    py2exe = None
else:
    class py2exe(_py2exe):
        def copy_w9xpopen(self, *args, **kwargs):
            pass

        def copy_dlls(self, dlls):
            _py2exe.copy_dlls(self, dlls)

            if self.bundle_files < 3:
                raise AssertionError(
                  "can't include Qt plugins if bundle_files < 3"
                )

            import PyQt4

            src_dir = os.path.join(
              os.path.dirname(PyQt4.__file__),
              'plugins',
              'imageformats',
            )

            dst_dir = os.path.join(self.exe_dir, 'imageformats')
            self.mkpath(dst_dir)

            for src_basename in os.listdir(src_dir):
                src = os.path.join(src_dir, src_basename)
                dst = os.path.join(dst_dir, src_basename)
                _, copied = self.copy_file(src, dst, preserve_mode = 0)
                self.lib_files.append(dst)


    extra_kwargs['cmdclass']['py2exe'] = py2exe


###


VCREDIST_URL = 'http://download.microsoft.com/download/1/1/1/1116b75a-9ec3-481a-a3c8-1777b5381140/vcredist_x86.exe'


class makensis(Command):
    description = 'build Windows installer'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        py2exe_obj = self.get_finalized_command('py2exe')
        py2exe_obj.run()

        if not os.path.exists('misc/vcredist_x86.exe'):
            # At least I think this is true.
            log.info(
              'WARNING: '
              'You must have a Visual Studio license to '
              'distribute the installer.'
            )
            log.info('downloading vcredist_x86.exe...')
            try:
                with open('misc/vcredist_x86.exe', 'wb') as f:
                    src = urlopen(VCREDIST_URL)
                    try:
                        while True:
                            data = src.read(1000)
                            if not data:
                                break
                            f.write(data)
                    finally:
                        src.close()
            except:
                # Don't leave partial files lying around.
                os.unlink('misc/vcredist_x86.exe')
                raise

        check_call(['makensis', 'rapidwriter.nsi'], shell = True)


extra_kwargs['cmdclass']['makensis'] = makensis


###


def get_version(release_file):
    try:
        f = open(release_file, 'r')
        try:
            return f.read().strip()
        finally:
            f.close()
    except (IOError, OSError):
        status, output = commands.getstatusoutput('bzr revno')
        if status == 0:
            return 'bzr%s' % output.strip()
    return 'unknown'


if sys.platform == 'win32':
    import win32api, win32con

    RT_MANIFEST = 24

    h = win32api.LoadLibraryEx(
      sys.executable,
      None,
      win32con.LOAD_LIBRARY_AS_DATAFILE,
    )
    try:
        resource_names = win32api.EnumResourceNames(h, RT_MANIFEST)
        if len(resource_names) != 1:
            raise AssertionError(resource_names)
        manifest_content = str(win32api.LoadResource(
          h,
          RT_MANIFEST,
          resource_names[0],
        ))
    finally:
        win32api.FreeLibrary(h)

    # rapidwriter requires administrative privileges.
    manifest_content = manifest_content.replace(
      'asInvoker',
      'requireAdministrator',
    )

    extra_kwargs.update({
      'options': {
        'py2exe': {
          # See http://stackoverflow.com/questions/1979486/py2exe-win32api-pyc-importerror-dll-load-failed
          'dll_excludes': ['mswsock.dll', 'powrprof.dll'],
          'includes' : ['sip'],
          'ignores': ['pyudev', 'pyudev.pyqt4'],
          'bundle_files': 3,
          'dist_dir': 'build/bdist.py2exe',
        },
      },
      'zipfile': None,
      'windows': [
        {
          'script' : 'rapidwriter',
          'icon_resources' : [(0, 'ui/resources/app_icon.ico')],
          'other_resources': [(24, 1, manifest_content)]
        },
      ],
      'data_files_arch_pkg': [
        'modules/rapidwriter/data/mbr.bin',
        'modules/rapidwriter/data/syslinux.exe',
      ],
      'data_files_indep_pkg': ['COPYING'],
    })

else:
    extra_kwargs.update({
      'data_files_indep': [
        ('applications', ['misc/rapidwriter.desktop']),
        ('pixmaps', ['misc/rapidwriter.png']),
      ],
    })


setup(
  name = 'rapidwriter',
  distinfo_module = 'rapidwriter.distinfo',
  version = get_version('release'),
  packages = [
    'rapidwriter',
    'rapidwriter.backends',
    'rapidwriter.backends.linux',
  ],
  package_dir = {'rapidwriter': os.path.join('modules', 'rapidwriter')},
  scripts = ['rapidwriter'],
  license = 'GNU General Public License (GPL)',
  platforms = ['Linux', 'Windows'],
  maintainer = 'Forest Bond',
  maintainer_email = 'forest.bond@rapidrollout.com',
  **extra_kwargs
)
